package io.tswf.transmutations.transmute.injection

import groovy.transform.CompileStatic
import io.tswf.transmutations.transmute.exception.InvalidTransmutationException
import io.tswf.transmutations.transmute.injection.contexts.TransmutationInjectionToMethodAtInsnContext
import io.tswf.transmutations.transmute.injection.hook.DefaultHookConstants
import io.tswf.transmutations.transmute.meta.TransmutationMethodMetadata
import org.objectweb.asm.tree.*

import java.lang.reflect.Modifier

import static org.objectweb.asm.Opcodes.*

@CompileStatic
class TransmutationMethodAtInsnInjector {

    public static int transferredMethodNameLength = 20

    static void injectTransmutation(
            TransmutationInjectionToMethodAtInsnContext insnContext
    ) {
        def transmutationMeta = insnContext.allPointsContext.transmutationMetadata
        def targetMethod = insnContext.allPointsContext.targetMethod
        def targetMethodIsStatic = Modifier.isStatic(targetMethod.access)

        validateTransmutationAndTarget(transmutationMeta, targetMethod)

        // Потому как в контексте коммиты происходят в режиме AFTER, то чтобы вставлять хуки перед инструкцией-целью требуется сделать шаг назад по-возможности.
        if (insnContext.currentInsn == null)
            insnContext.currentInsn = insnContext.startInsn.previous ?: insnContext.startInsn //InjectionInsnOffsetUtil.applyInsnOffset(targetMethod, insnContext.startInsn, insnContext.injectionOffset)

        def transmutationCallMethod = transmutationMeta.transmutationMethod.asAsmMethod()
        def transmutationCallOwner = transmutationMeta.transmutationOwner.asAsmType()

        // Обрабатываем хуки, которые должны срабатывать перед инъекцией.
        // Например тут могут создаваться дополнительные лейблы, к которым потом можно будет прыгать по всему методу.
        for (def preInjectionHook : transmutationMeta.preInjectionHooks) {
            preInjectionHook.handleTransmutationPreInjection(insnContext)
            insnContext.commit()
        }

        // Начинаем вызов трансмутационного метода

        // Кладем на стек this, если метод трансмутации не-статический
        if (!Modifier.isStatic(transmutationMeta.transmutationMethod.access)) {
            insnContext.insnsToInsertToCurrent.add(new VarInsnNode(ALOAD, 0))
        }

        // Загружаем все параметры метода на стек
        def loadedParametersCount = 0
        def loadingParamIndex = targetMethodIsStatic ? 0 : 1
        for (def targetMethodParameterType : targetMethod.parameterTypes) {
            insnContext.insnsToInsertToCurrent.add(new VarInsnNode(targetMethodParameterType.getOpcode(ILOAD), loadingParamIndex++))
            loadedParametersCount ++
        }

        // Загружаем кастомные параметры, будь то выброшенное исключение или же локальная переменная.
        // Загрузчики уже расположены в том же порядке, в каком расположены они в методе трансмутации.
        for (def customParameterLoader : transmutationMeta.customParameterLoaders) {
            customParameterLoader.loadCustomParameter(insnContext, loadedParametersCount, transmutationMeta.transmutationMethod.getParameterType(loadedParametersCount))
            loadedParametersCount ++
        }

        if (loadedParametersCount < transmutationMeta.transmutationMethod.parameterTypes.size()) {
            //TODO: Сделать тут более информативное и красивое сообщение, с указанием подробной информации о том, для каких параметров нет загрузчика
            throw new InvalidTransmutationException("There is not all parameters loaded for transmutation method ${transmutationMeta.transmutationMethod.javaFullName} of ${transmutationMeta.transmutationOwner.javaName}")
        }

        // Когда загружены все параметры - вызываем метод трансмутации
        insnContext.insnsToInsertToCurrent.add(
                new MethodInsnNode(
                        Modifier.isStatic(transmutationMeta.transmutationMethod.access)
                                ? INVOKESTATIC
                                : INVOKEVIRTUAL,
                        transmutationCallOwner.internalName,
                        transmutationCallMethod.name,
                        transmutationCallMethod.descriptor,
                        false
                )
        )

        // Выгружаем полученные инструкции в тело метода-цели и очищаем буфер
        insnContext.commit()

//        // Указываем после какой ноды должны генерировать инструкции хуки пост-инъекции
//        insnContext.currentInsn = insnBeforeWhichInsert

        // Обработка возвращенного из трансмутационного метода значения
        transmutationMeta.returnValueProcessor.handleTransmutationMethodReturnValue(insnContext)
        insnContext.commit()

        // Выполняем хуки пост-инъекции
        for (def postInjectHook : transmutationMeta.postInjectionHooks) {
            postInjectHook.handleTransmutationPostInjection(insnContext)
            insnContext.commit()
        }

        if (insnContext.customPayload.getByKey(DefaultHookConstants.SKIP_EXIT_INSN)) {
            targetMethod.instructions.remove(insnContext.startInsn)
        }
    }

    static void validateTransmutationAndTarget(
        TransmutationMethodMetadata transmutationMethodMetadata,
        MethodNode targetMethod
    ) {
        def transmutationMethod = transmutationMethodMetadata.transmutationMethod
        def transmutationIsStatic = Modifier.isStatic(transmutationMethod.access)

        if (!transmutationIsStatic) {
            throw new InvalidTransmutationException("The transmutation method ${transmutationMethod.javaFullName} of ${transmutationMethodMetadata.transmutationOwner.javaName} must be static!")
        }
    }
}
