package io.tswf.transmutations.transmute.injection.returnvalue

import groovy.transform.CompileStatic
import io.tswf.transmutations.transmute.exception.InvalidTransmutationException
import io.tswf.transmutations.transmute.injection.ITransmutationMethodReturnValueProcessor
import io.tswf.transmutations.transmute.injection.contexts.TransmutationInjectionToMethodAtInsnContext
import io.tswf.transmutations.transmute.injection.hook.DefaultHookConstants
import org.objectweb.asm.Type
import org.objectweb.asm.tree.InsnNode
import org.objectweb.asm.tree.JumpInsnNode
import org.objectweb.asm.tree.LabelNode

import static io.tswf.groovy.bettergroovy.asm.util.BoxingUtil.instance as boxing
import static org.objectweb.asm.Opcodes.DUP
import static org.objectweb.asm.Opcodes.IFNULL
import static org.objectweb.asm.Opcodes.IRETURN
import static org.objectweb.asm.Opcodes.POP

@CompileStatic
class ReturnObjectIfNotNullReturnValueProcessor implements ITransmutationMethodReturnValueProcessor {

    @Override
    void handleTransmutationMethodReturnValue(TransmutationInjectionToMethodAtInsnContext context) {
        def targetMethod = context.allPointsContext.targetMethod
        def transmutationMethod = context.allPointsContext.transmutationMetadata.transmutationMethod

        def targetMethodReturnType = targetMethod.returnType
        def transmutationMethodReturnType = transmutationMethod.returnType

        //TODO: Может быть сделать более умное сравнение?

        // Учитываем случай, когда трансмутация возвращает Integer, а целевой метод - int.
        if (boxing.getBoxedDescriptorOrSelf(targetMethodReturnType.descriptor) != transmutationMethodReturnType.descriptor) {
            throw new InvalidTransmutationException("Target method must has same return type with transmutation! expected: ${transmutationMethodReturnType.descriptor}, but found ${targetMethodReturnType.descriptor}")
        }

        // Если метод трансмутации имеет не объектный тип (а есть примитив или void), то возвращаемый из него объект нельзя проверить на null. Следовательно ошибка.
        if (transmutationMethodReturnType.isPrimitive() || transmutationMethodReturnType.descriptor == Type.VOID_TYPE.descriptor) {
            throw new InvalidTransmutationException("Transmutation method ${transmutationMethod.javaFullName} of ${context.allPointsContext.transmutationMetadata.transmutationOwner.javaName} must have non-primitive return type!")
        }

        def nullLabel = new LabelNode()
        context.insnsToInsertToCurrent << new InsnNode(DUP)
        context.insnsToInsertToCurrent << new JumpInsnNode(IFNULL, nullLabel)

        // Если возвращаемые типы не одинаковые, значит трансмутация имеет оберточный тип и требует развертки в примитив
        if (targetMethodReturnType.descriptor != transmutationMethodReturnType.descriptor) {

            // Оберткой в этом случае может быть только возвращаемый из метода трансмутации объект, так что по его типу и делаем анбоксинг
            boxing.applyUnboxing(transmutationMethodReturnType.descriptor, context.insnsToInsertToCurrent)
        }

        context.insnsToInsertToCurrent << new InsnNode(targetMethodReturnType.getOpcode(IRETURN))

        context.insnsToInsertToCurrent << nullLabel
        context.insnsToInsertToCurrent << new InsnNode(POP)
    }
}
